"""Class definition for DebPackageData."""

from __future__ import annotations

from typing import TYPE_CHECKING

from debian.debian_support import Version
from hoppr import Component

from choppr.apt_api import BinaryPackage
from choppr.types.choppr_shares import ChopprShares
from choppr.types.package_data import PackageData
from choppr.utils import filenames_match


if TYPE_CHECKING:
    from choppr.apt_api.dependency import Dependency


__all__ = ["DebPackageData"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class DebPackageData(PackageData):
    """Class containing frequently accessed information and the originating package.

    Members:
        - name
        - package
        - version
    """

    def __init__(self, package: BinaryPackage) -> None:
        """Create an instance of DebPackageData.

        Arguments:
            package: The originating binary package
        """
        self.package: BinaryPackage = package
        self.name: str = package.package
        self.version: str = package.version

    def provides_file(self, file: str) -> bool:
        """Check if the package provides the given file.

        Arguments:
            file: The file to search for

        Returns:
            bool: True if the file is provided by this package
        """
        return any(
            filenames_match(provided_file, file, force_partial_match=True) for provided_file in self.package.provides
        )

    def satisfies_dependency(self, dependency: Dependency) -> bool:
        """Check if the package satisfies the given dependency.

        Arguments:
            dependency: The dependency to check against.

        Returns:
            bool: True if the depencency is satisfied by this package
        """
        return self.name == dependency.name and self._version_match(dependency.version)

    def is_essential(self, build_essential: bool = False) -> bool:
        """Check if the package is essential.

        Arguments:
            build_essential: Enable checking build essential packages (default False)

        Returns:
            bool: True if package is essential
        """
        return bool(
            self.package.essential
            or (build_essential and self.package.build_essential)
            or (self.package.priority and self.package.priority.lower() == "required")
        )

    @staticmethod
    def _get_min_max_version(versions: list[str], symbol: str) -> tuple[bool, Version | None]:
        return next(
            (("=" in v, Version(v.replace(symbol, "").replace("=", "").strip())) for v in versions if symbol in v),
            (False, None),
        )

    def _get_max_version(self, versions: list[str]) -> tuple[bool, Version | None]:
        return self._get_min_max_version(versions, "<")

    def _get_min_version(self, versions: list[str]) -> tuple[bool, Version | None]:
        return self._get_min_max_version(versions, ">")

    def _version_within_max(self, versions: list[str]) -> bool:
        can_be_equal, max_version = self._get_max_version(versions)

        if max_version is None:
            ChopprShares.log.debug(f"No valid maximum version: {', '.join(versions)}")
            return False

        return (can_be_equal and Version(self.package.version) <= max_version) or (
            not can_be_equal and Version(self.package.version) < max_version
        )

    def _version_within_min(self, versions: list[str]) -> bool:
        can_be_equal, min_version = self._get_min_version(versions)

        if min_version is None:
            ChopprShares.log.debug(f"No valid minimum version: {', '.join(versions)}")
            return False

        return (can_be_equal and Version(self.package.version) >= min_version) or (
            not can_be_equal and Version(self.package.version) > min_version
        )

    def _version_within_range(self, versions: list[str]) -> bool:
        within_max = self._version_within_max(versions)
        within_min = self._version_within_min(versions)

        return within_max and within_min

    def _version_match(self, version: str) -> bool:
        # sourcery skip: assign-if-exp, reintroduce-else
        if version.startswith("="):
            exact_version = Version(version.replace("=", "").strip())
            return Version(self.package.version) == exact_version
        if all(sign in version for sign in ["<", ">"]):
            return self._version_within_range(version.split(","))
        if version.startswith("<"):
            return self._version_within_max([version])
        if version.startswith(">"):
            return self._version_within_min([version])

        return True

    def __eq__(self, other: object) -> bool:
        match other:
            case DebPackageData():
                return self.name == other.name and self.version == other.version
            case BinaryPackage():
                return self.name == str(other.package) and self.version == str(other.version)
            case Component():
                if ChopprShares.config.allow_version_mismatch:
                    return self.name == other.name
                if other.version is not None:
                    return self.name == other.name and self.version == other.version

        return False

    __hash__ = PackageData.__hash__



"""Choppr refines the components in a Software Bill of Materials (SBOM).

It does not replace SBOM generation tools.  Mainly, Choppr analyses a build or runtime to verify
which components are used, and remove the SBOM components not used.  Starting with file accesses, it
works backwards from how an SBOM generation tool typically would.  For example SBOM generators use
the yum database to determine which packages yum installed.  Choppr looks at all the files accessed
and queries sources like yum to determine the originating package.
"""

from __future__ import annotations

import re
import shutil

from threading import Lock, Thread
from typing import TYPE_CHECKING, Any

from hoppr import (
    BomAccess,
    ComponentType,
    HopprLoadDataError,
)
from hoppr.base_plugins.hoppr import HopprPlugin, hoppr_process
from hoppr.models.types import PurlType
from hoppr.result import Result
from hoppr_cyclonedx_models.cyclonedx_1_6 import Scope

from choppr import __version__
from choppr.component_handlers.deb_handler import DebHandler
from choppr.component_handlers.rpm_handler import RpmHandler
from choppr.types import ChopprConfig, ChopprConfigModel, ChopprShares
from choppr.types.choppr_config import OperatingMode
from choppr.utils import compress_directory, extract_archive, get_purl_type, log_header, output_list


if TYPE_CHECKING:
    from hoppr import Component, HopprContext, Sbom

    from choppr.component_handlers.purl_handler import PurlHandler


__all__ = ["Choppr"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _clear_cache() -> None:
    ChopprShares.log.info("Clearing cache")
    shutil.rmtree(ChopprShares.config.cache_dir)


class Choppr(HopprPlugin):
    """Plugin implementation of Choppr to integrate with Hoppr."""

    supported_purl_types = ["rpm", "deb"]  # noqa: RUF012
    products = []  # noqa: RUF012
    bom_access = BomAccess.FULL_ACCESS

    def __init__(self, context: HopprContext, config: dict[Any, Any] | None = None) -> None:
        """Initialize plugin with Hoppr framework arguments (context and config).

        Arguments:
            context: Hoppr context
            config: Hoppr configuration (default None)
        """
        super().__init__(context, config)

        self.log = self.get_logger()
        self.log.flush_immed = True

        # Parse configuration
        self.valid_config = False
        try:
            plugin_config = ChopprConfig(ChopprConfigModel.parse_obj(config))
            ChopprShares(plugin_config, self.context, self.log)
            self.valid_config = True
        except (HopprLoadDataError, FileNotFoundError) as e:
            self.log.error(f"Invalid Configuration: {e}")  # noqa: TRY400

        self.search_repositories = dict.fromkeys(PurlType, False)
        self.preprocessed_files: set[str] = set()
        self.preprocessed_components: set[Component] = set()
        self.purl_handlers: dict[PurlType, PurlHandler] = {PurlType.RPM: RpmHandler(), PurlType.DEB: DebHandler()}
        self.unimplemented_purl_types_logged: set[PurlType] = set()
        self._lock = Lock()

    def get_version(self) -> str:  # noqa: PLR6301
        """Return the version of the Choppr plugin.

        Returns:
            str: Plugin version
        """
        return __version__

    def _run_handler_methods_concurrently(
        self,
        method_name: str,
        *args: Any,  # noqa: ANN401
        purl_types: set[PurlType] | None = None,
        **kwargs: Any,  # noqa: ANN401
    ) -> tuple[list[Thread], dict[PurlType, Any]]:
        if purl_types is None:
            purl_types = set(self.purl_handlers.keys())

        results: dict[PurlType, Any] = dict.fromkeys(purl_types)
        threads: list[Thread] = []

        def _run_with_result(handler: PurlHandler, purl_type: PurlType) -> None:
            """Run the method and store its return value."""
            method = getattr(handler, method_name)
            result = method(*args, **kwargs)

            with self._lock:
                results[purl_type] = result

        # Create threads for each handler
        for purl_type in purl_types:
            if purl_type in self.purl_handlers:
                thread = Thread(target=_run_with_result, args=(self.purl_handlers[purl_type], purl_type))
                threads.append(thread)
                thread.start()

        return threads, results

    ################################################################################################
    # Pre-Stage Process
    ################################################################################################

    @hoppr_process
    def pre_stage_process(self) -> Result:
        """Collect repository information.

        Returns:
            Result: Skip if the config is invalid, or the modified SBOM
        """
        if not self.valid_config:
            return Result.skip()

        choppd_sbom = None

        match ChopprShares.config.mode:
            case OperatingMode.CACHE:
                self._cache_repositories(True)
            case OperatingMode.RUN:
                if ChopprShares.config.cache_input:  # Extract provided cache
                    extract_archive(ChopprShares.config.cache_input, ChopprShares.config.cache_dir)
                self._pre_process_components()
                self._cache_repositories()
                self._resolve_component_packages()
                self._populate_required_components()
                self._populate_dependency_components()

                choppd_sbom = self._filter_sbom()

                self._output_excluded_components(choppd_sbom.components)

        return Result.success(return_obj=choppd_sbom)

    def _pre_process_components(self) -> None:
        log_header("Pre-Processing Components")

        for file in ChopprShares.config.strace_files:
            for pattern, component in ChopprShares.pattern_components:
                if re.search(pattern, file):
                    ChopprShares.log.debug(f'File "{file}" provided by {component.name}-{component.version}')
                    self.preprocessed_files.add(file)
                    self.preprocessed_components.add(component)

    def _cache_repositories(self, clear: bool = False) -> None:
        log_header("Cache Repositories")

        if clear:
            _clear_cache()

        threads, results = self._run_handler_methods_concurrently("cache_repositories")
        for thread in threads:
            thread.join()

        # Update search_repositories with return values
        with self._lock:
            for purl_type, result in results.items():
                if isinstance(result, bool):
                    self.search_repositories[purl_type] = result

    def _resolve_component_packages(self) -> None:
        log_header("Resolve Component Packages")

        if purl_types := {pt for pt in self.purl_handlers if self.search_repositories[pt]}:
            threads, _ = self._run_handler_methods_concurrently("resolve_component_packages", purl_types=purl_types)
            for thread in threads:
                thread.join()

    def _populate_required_components(self) -> None:
        log_header("Populate Required Components")
        required_files = ChopprShares.config.strace_files - self.preprocessed_files

        if purl_types := {pt for pt in self.purl_handlers if self.search_repositories[pt]}:
            threads, _ = self._run_handler_methods_concurrently(
                "populate_required_components", required_files, purl_types=purl_types
            )
            for thread in threads:
                thread.join()

    def _populate_dependency_components(self) -> None:
        log_header("Populate Dependency Components")

        if purl_types := {pt for pt in self.purl_handlers if self.search_repositories[pt]}:
            threads, _ = self._run_handler_methods_concurrently("populate_dependency_components", purl_types=purl_types)
            for thread in threads:
                thread.join()

    @staticmethod
    def _allowlist_denylist_scope(component: Component) -> Scope | None:
        component_id = f"{component.name}-{component.version}"

        if purl_type := get_purl_type(component):
            if purl_type in ChopprShares.config.allowlist and component in list(
                ChopprShares.config.allowlist[purl_type]
            ):
                ChopprShares.log.debug(f"Component accepted by allowlist: {component_id}")
                return Scope.REQUIRED
            if purl_type in ChopprShares.config.denylist and component in list(ChopprShares.config.denylist[purl_type]):
                ChopprShares.log.debug(f"Component blocked by denylist: {component_id}")
                return Scope.EXCLUDED

        return None

    def _filter_sbom(self) -> Sbom:
        log_header("Filter SBOM")
        choppd_sbom = ChopprShares.context.delivered_sbom.copy(deep=True)
        choppd_sbom.components.clear()
        components_required = 0
        components_excluded = 0
        components_unknown = 0

        for component in ChopprShares.context.delivered_sbom.components or []:
            component_id = f"{component.name}-{component.version}"

            # Previously Excluded
            if component.scope == Scope.EXCLUDED:
                ChopprShares.log.debug(f"Component previously excluded: {component_id}")
                components_excluded += 1
                continue

            # Allowlist/Denylist
            match scope := self._allowlist_denylist_scope(component):
                case Scope.REQUIRED:
                    component.scope = scope
                    choppd_sbom.components.append(component)
                    components_required += 1
                    continue
                case Scope.EXCLUDED:
                    component.scope = scope
                    components_excluded += 1
                    continue

            # Pre-Processed Components
            if component in self.preprocessed_components:
                ChopprShares.log.debug(f"Component required: {component_id}")
                component.scope = Scope.REQUIRED
                choppd_sbom.components.append(component)
                components_required += 1
                continue

            # Component Parsing
            match component_scope := self._get_component_scope(component):
                case Scope.REQUIRED:
                    ChopprShares.log.debug(f"Component required: {component_id}")
                    component.scope = component_scope
                    choppd_sbom.components.append(component)
                    components_required += 1
                    continue
                case Scope.EXCLUDED:
                    ChopprShares.log.debug(f"Component not required: {component_id}")
                    component.scope = component_scope
                    components_excluded += 1
                    continue

            components_unknown += 1

        ChopprShares.log.info(f"Components Required: {components_required}")
        ChopprShares.log.info(f"Components Excluded: {components_excluded}")
        ChopprShares.log.info(f"Components Unknown: {components_unknown}")

        if ChopprShares.config.delete_excluded:
            ChopprShares.log.info("Deleted excluded components")
            return choppd_sbom
        return ChopprShares.context.delivered_sbom

    @staticmethod
    def _output_excluded_components(filtered_components: list[Component]) -> None:
        excluded_components_all = [
            c for c in ChopprShares.context.delivered_sbom.components if c not in filtered_components
        ]

        for purl_type in PurlType:
            output_format = ChopprShares.config.output_files.excluded_components[purl_type].component_format
            if excluded_components := [
                output_format.format(**c.dict()) for c in excluded_components_all if get_purl_type(c) is purl_type
            ]:
                output_file = ChopprShares.config.output_files.excluded_components[purl_type].file
                output_file.parent.mkdir(parents=True, exist_ok=True)
                output_list(output_file, excluded_components)

    def _get_component_scope(self, component: Component) -> Scope | None:
        scope: Scope | None = None

        if (
            ChopprShares.config.keep_essential_os_components
            and str(component.type) == ComponentType.OPERATING_SYSTEM.value
        ):
            scope = Scope.REQUIRED
        elif not ChopprShares.config.allow_version_mismatch and not component.version:
            pass
        elif (
            str(component.type) == ComponentType.FILE.value and component.name in ChopprShares.config.strace_files
        ) or component in self._get_purl_required_components() | self._get_purl_dependency_components():
            scope = Scope.REQUIRED
        elif (purl_type := get_purl_type(component)) in self.purl_handlers:
            scope = Scope.EXCLUDED
        elif purl_type and purl_type not in self.unimplemented_purl_types_logged:
            ChopprShares.log.debug(f"Purl support not implemented yet: {purl_type}")
        else:
            ChopprShares.log.debug(f"Unsupported component type in SBOM: {component.type}")

        return scope

    def _get_purl_required_components(self) -> set[Component]:
        return {component for handler in self.purl_handlers.values() for component in handler.required_components}

    def _get_purl_dependency_components(self) -> set[Component]:
        return {component for handler in self.purl_handlers.values() for component in handler.dependency_components}

    ################################################################################################
    # Post-Stage Process
    ################################################################################################

    @hoppr_process
    def post_stage_process(self) -> Result:
        """Perform cleanup tasks after running Choppr.

        Returns:
            Result: The updated SBOM with the unused packages removed
        """
        if not self.valid_config:
            return Result.skip()

        if ChopprShares.config.mode == OperatingMode.CACHE or ChopprShares.config.archive_cache:
            ChopprShares.log.info("Creating cache archive...")
            compress_directory(ChopprShares.config.output_files.cache_archive, ChopprShares.config.cache_dir)
            ChopprShares.log.info(f"Cache archive written to {ChopprShares.config.output_files.cache_archive}")

        if ChopprShares.config.clear_cache:
            _clear_cache()

        return Result.success()



"""PackagesFile implementation."""

from __future__ import annotations

import json

from typing import TYPE_CHECKING

from hoppr import PurlType

from choppr.apt_api._utils import get_value
from choppr.apt_api.binary_package import BinaryPackage
from choppr.types.choppr_config import OperatingMode
from choppr.types.choppr_shares import ChopprShares


if TYPE_CHECKING:
    from pathlib import Path


__all__ = ["PackagesFile"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _get_file_list(binary_package: BinaryPackage) -> list[str]:
    ChopprShares.log.debug(f"Package not found in contents file, getting provided files: {binary_package.package}")
    try:
        if package_file_list := binary_package.provides:
            return package_file_list
    except ConnectionError:
        ChopprShares.log.warning(f"Failed to get file list for {binary_package.package}, continuing...")

    return []


class PackagesFile:
    """Class that represents a Packages file."""

    def __init__(self, packages_file: Path, content_file: Path, repo_url: str) -> None:
        """Initialize an instance of PackagesFile.

        Arguments:
            packages_file: The list of packages with their metadata
            content_file: The list of files internal to this specific package
            repo_url: The URL for the repository
        """
        self.packages: list[BinaryPackage] = []

        if packages_file.is_file():
            with packages_file.open(encoding="utf-16") as f:
                content = f.read().strip()

            with content_file.open(encoding="utf-8") as f:
                file_list: dict[str, list[str]] = json.load(f)

            file_list_updated = False

            if packages_data := content.split("\n\n"):
                for package_data in [p for p in packages_data if p]:
                    name = str(get_value(package_data, "Package"))
                    if name in file_list:
                        bp = BinaryPackage(package_data, file_list[name], repo_url)
                    else:
                        bp = BinaryPackage(package_data, package_url=repo_url)

                        if (
                            ChopprShares.config.mode == OperatingMode.CACHE
                            and any(c for c in ChopprShares.purl_components[PurlType.DEB] if c.name == name)
                            and (package_file_list := _get_file_list(bp))
                        ):
                            file_list[name] = package_file_list
                            file_list_updated = True

                    self.packages.append(bp)

            if file_list_updated:
                with content_file.open("w", encoding="utf-8") as f:
                    json.dump(file_list, f, cls=json.JSONEncoder, indent=2)



"""Class definition for ChopprShares, a singleton class to share commonly access objects."""

from __future__ import annotations

import re

from typing import TYPE_CHECKING

from hoppr import ComponentType, PurlType


if TYPE_CHECKING:
    from hoppr import Component, HopprContext, HopprLogger
    from typing_extensions import Self

    from choppr.types import ChopprConfig


__all__ = ["ChopprShares"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _has_choppr_pattern(component: Component) -> bool:
    return "choppr-pattern" in {prop.name for prop in component.properties}


def _get_choppr_pattern(component: Component) -> re.Pattern[str] | None:
    choppr_pattern = next((prop.value for prop in component.properties if prop.name == "choppr-pattern"), None)
    return re.compile(choppr_pattern) if choppr_pattern else None


class ChopprShares:
    """Singleton class to hold commonly accessed Choppr objects.

    Members:
        - components
        - config
        - context
        - log

    Methods:
        - initialized
    """

    _instance: Self | None = None

    def __new__(cls, config: ChopprConfig, context: HopprContext, log: HopprLogger) -> Self:
        """Initialize ChopprShares singleton.

        Arguments:
            config: Instance of ChopprConfig
            context: Instance of HopprContext
            log: Instance of HopprLogger

        Returns:
            Self: Singleton instance of ChopprShares
        """
        if cls._instance is None:
            cls._instance = super().__new__(cls)

            cls.pattern_components: set[tuple[re.Pattern[str], Component]] = {
                (pattern, component)
                for component in context.delivered_sbom.components
                if _has_choppr_pattern(component) and (pattern := _get_choppr_pattern(component)) is not None
            }
            cls.purl_components: dict[PurlType, list[Component]] = {
                purl_type: [
                    component
                    for component in context.delivered_sbom.components
                    if component.purl and component.purl.lower().startswith(f"pkg:{purl_type.value.lower()}")
                ]
                for purl_type in PurlType
            }
            cls.file_components: set[Component] = {
                component
                for component in context.delivered_sbom.components
                if str(component.type) == ComponentType.FILE.value
            }
            cls.config: ChopprConfig = config
            cls.context: HopprContext = context
            cls.log: HopprLogger = log

            cls.log.debug("Purl component counts:")
            for purl_type, component_list in cls.purl_components.items():
                if component_count := len(component_list):
                    cls.log.debug(f"{purl_type.name}: {component_count}", indent_level=1)
            cls.log.debug(f"File component count: {len(cls.file_components)}")
        return cls._instance

    @classmethod
    def initialized(cls) -> bool:
        """Check if the class has been initialized.

        Returns:
            bool: True if the class is initialized, otherwise False
        """
        return cls._instance is not None



"""Module containing functions to call strace and analyze output."""

from __future__ import annotations

import re

from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from pathlib import Path


__all__ = ["get_files"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "Lockheed Martin Proprietary Information"


def _resolve_relative_path(path: str) -> str:
    absolute_path = path
    relative_path_pattern = r"[\/|\\][^\/|\\]+?[\/|\\]\.\."

    while re.search(relative_path_pattern, absolute_path):
        absolute_path = re.sub(relative_path_pattern, "", absolute_path)

    return absolute_path


def _remove_repeated_slashes(path: str) -> str:
    slash = r"\\" if r"\\" in path else "/"
    return re.sub(r"[\/|\\]{2,}", slash, path)


def _clean_path(path: str) -> str:
    clean_path = _resolve_relative_path(path)
    return _remove_repeated_slashes(clean_path)


def get_files(strace_location: Path) -> set[str]:
    """Get the list of touched files from strace.

    Arguments:
        strace_location: Output file from strace

    Returns:
        set[str]: Set of touched files
    """
    with strace_location.open(encoding="utf-8") as file:
        text = file.read()

    flags = re.MULTILINE | re.IGNORECASE

    dir_regex = (
        r"^.*?"
        r"(?:(?:mkdir).*?"
        r"(?:(?:\'|\")(?P<mkdir>.*?)(?:\'|\")))"
        r"|"
        r"(?:(?:(?:\'|\")(?P<dirname>.*?)(?:\'|\"))"
        r"(?=.*?(?:S_IFDIR|O_DIRECTORY))(?!.*(?:ENOENT|unfinished)))"
    )
    dir_matches = re.finditer(dir_regex, text, flags)
    directories: set[str] = {
        _clean_path(dm.group(g)) for dm in dir_matches for g in ["dirname", "mkdir"] if dm.group(g) is not None
    }

    regex = r"^.*?(?:(?:\'|\")(?P<filename>.*?)(?:\'|\"))(?!.*(?:S_IFDIR|ENOENT|O_DIRECTORY|unfinished))"
    files: set[str] = {_clean_path(match.group("filename")) for match in re.finditer(regex, text, flags)}
    return files - directories



"""Class definition for PackageData."""

from __future__ import annotations

import abc

from hashlib import sha256
from typing import Any

from hoppr import Component

from choppr.types.choppr_shares import ChopprShares


__all__ = ["PackageData"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2025 Lockheed Martin Corporation"
__license__ = "MIT License"


class PackageData(abc.ABC):
    """Abstract base class for package data, providing common functionality for package metadata."""

    def __init__(self, package: Any) -> None:  # noqa: ANN401
        """Initialize the PackageData with the originating package.

        Arguments:
            package: The originating package data (e.g., dictionary for RPM, BinaryPackage for Debian)
        """
        self.package: Any = package
        self.name: str = ""
        self.version: str = ""

    @abc.abstractmethod
    def provides_file(self, file: str) -> bool:
        """Check if the package provides the given file.

        Arguments:
            file: The file to search for

        Returns:
            bool: True if the file is provided by this package
        """

    def __eq__(self, other: object) -> bool:
        match other:
            case PackageData():
                return self.name == other.name and self.version == other.version
            case Component():
                if ChopprShares.config.allow_version_mismatch:
                    return self.name == other.name
                if other.version is not None:
                    return self.name == other.name and self.version == other.version

        return False

    def __hash__(self) -> int:
        sha = sha256()
        sha.update(self.name.encode())
        sha.update(self.version.encode())
        return int(sha.hexdigest(), 16)

    def __str__(self) -> str:
        return f"{self.name}-{self.version}"



"""Class definition for RpmPackageData."""

from __future__ import annotations

from pathlib import Path
from typing import Any

import jmespath

from hoppr import Component

from choppr.types.choppr_shares import ChopprShares
from choppr.types.package_data import PackageData
from choppr.utils import filenames_match, remove_parenthesis


__all__ = ["RpmPackageData"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class RpmPackageData(PackageData):
    """Class containing frequently accessed information and the originating package.

    Members:
        - name
        - package
        - short_version
        - release
        - version
    """

    def __init__(self, package: dict[str, Any]) -> None:
        """Create an instance of PackageDetails.

        Arguments:
            package: The originating package dictionary
        """
        self.package: dict[str, Any] = package
        self.name: str = package["name"]
        self.short_version: str = package["version"]["@ver"]
        self.release: str = package["version"]["@rel"]
        self.version: str = f"{self.short_version}-{self.release}"

    def provides_file(self, file: str, search_files: bool = False) -> bool:
        """Check if the package provides the given file.

        Arguments:
            file: The file to search for
            search_files: Flag to enable searching the files section (default False)

        Returns:
            bool: True if the file is provided by this package
        """
        # Search provides section
        if (
            package_provides := jmespath.search(
                expression=('format."rpm:provides"."rpm:entry"[*]."@name"'),
                data=self.package,
            )
        ) and any(filenames_match(remove_parenthesis(provides), Path(file).name) for provides in package_provides):
            return True

        if search_files:
            # Search files
            if jmespath.search(
                expression=(f"format.file[?type(@) == 'string' && @ == '{file}']"),
                data=self.package,
            ):
                return True

            # Search directories
            directories: list[str] | None = jmespath.search(
                expression='format."file"[?"@type" == \'dir\']."#text"',
                data=self.package,
            )
            if directories and any(Path(file).is_relative_to(directory) for directory in directories):
                return True

        return False

    def requires(self) -> set[str]:
        """Get the files that the package requires.

        Returns:
            set[str]: The required files
        """
        if requirements := jmespath.search(
            expression='format."rpm:requires"."rpm:entry"[*]."@name"',
            data=self.package,
        ):
            return {requirement for requirement in requirements if requirement}

        return set()

    def __eq__(self, other: object) -> bool:
        match other:
            case RpmPackageData():
                return self.name == other.name and self.version == other.version
            case Component():
                if ChopprShares.config.allow_version_mismatch:
                    return self.name == other.name
                if other.version is not None:
                    return self.name == other.name and self.version == other.version
            case {"name": _, "version": {"@ver": _}}:
                return self.name == str(other["name"]) and self.short_version == str(other["version"]["@ver"])  # type: ignore[index]
        return False

    __hash__ = PackageData.__hash__



"""Module for handling DEB packages."""

from __future__ import annotations

from typing import TYPE_CHECKING, Final

from hoppr.models.manifest import Repository
from hoppr.models.types import PurlType
from pydantic import HttpUrl, parse_obj_as

from choppr import apt_api
from choppr.component_handlers.purl_handler import PurlHandler
from choppr.decorators import limit_recursion
from choppr.types import ChopprShares, DebPackageData
from choppr.utils import get_auth_and_verify, get_component_dependencies, get_purl_type, log_repo_pulls


if TYPE_CHECKING:
    from pathlib import Path

    from hoppr import Component


__all__ = ["DebHandler"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class DebHandler(PurlHandler):
    """Class to handle all DEB repository processing."""

    def __init__(self) -> None:
        super().__init__(PurlType.DEB)

        self.repositories: apt_api.Sources | None = None
        self.component_packages: dict[Component, dict[DebPackageData, bool]] = {}

    ####################################################################################################
    # Exported Methods
    ####################################################################################################

    def cache_repositories(self) -> bool:
        """Pull all of the metadata for DEB repositories provided in the config.

        Returns:
            bool: True when any repository was successfully pulled
        """
        if not (
            deb_repository_urls := [
                str(repo.url)
                for repo in parse_obj_as(list[Repository], ChopprShares.context.repositories[self.purl_type])
            ]
        ):
            return False

        config_deb_repository_urls = [str(repo.url) for repo in ChopprShares.config.deb_repositories]

        if set(deb_repository_urls) != set(config_deb_repository_urls):
            ChopprShares.log.error(
                "DEB repository mismatch: All repository URLs in the `manifest.yml` file should also be in the "
                "`transfer.yml` file"
            )

            for url in deb_repository_urls:
                ChopprShares.log.debug(f"Manifest Repository URL: {url}")
            for url in config_deb_repository_urls:
                ChopprShares.log.debug(f"Transfer Repository URL: {url}")

            return False

        repositories: list[apt_api.Repository] = []
        expected = sum(len(repo.distributions) for repo in ChopprShares.config.deb_repositories)
        ChopprShares.log.info(f"Pulling {expected} DEB repositories...")

        repositories_dir: Final[Path] = self.cache_dir / "repositories"
        repositories_dir.mkdir(parents=True, exist_ok=True)

        for repo in ChopprShares.config.deb_repositories:
            repo_url = parse_obj_as(HttpUrl, repo.url)
            for distribution in repo.distributions:
                auth, verify = get_auth_and_verify(repo_url)
                repositories.append(
                    apt_api.Repository(
                        repo_url,
                        distribution.name,
                        distribution.components,
                        repositories_dir,
                        auth,
                        verify,
                    )
                )

        self.repositories = apt_api.Sources(repositories)

        log_repo_pulls(expected, len(repositories), self.purl_type)

        return bool(repositories)

    def resolve_component_packages(self) -> None:
        """Resolve all DEB components to the packages in the repositories."""
        ChopprShares.log.info(
            f"Resolving packages for {len(ChopprShares.purl_components[self.purl_type])} RPM components"
        )
        for component in ChopprShares.purl_components[self.purl_type]:
            component_id = f"{component.name}-{component.version}"
            if packages := self._get_component_packages(component):
                if len(packages) == 1:
                    ChopprShares.log.debug(f"Resolved the package for {component_id}")
                else:
                    ChopprShares.log.warning(f"Resolved {len(packages)} packages for {component_id}")
                    for idx, package in enumerate(packages):
                        ChopprShares.log.warning(f"{idx}) {package}", indent_level=1)
            else:
                ChopprShares.log.warning(f"Unable to resolve a package for {component_id}")

            self.component_packages[component] = dict.fromkeys(packages, False)

    def populate_required_components(self, files: set[str]) -> None:
        """Determine which components provide the given files.

        Arguments:
            files: The files to search for in the component packages
        """
        required_packages: set[DebPackageData] = set()

        for file in files:
            file_found = False
            for component in ChopprShares.purl_components[self.purl_type]:
                if any(
                    (ChopprShares.config.keep_essential_os_components and package.is_essential())
                    or package.provides_file(file)
                    for package in self.component_packages[component]
                ):
                    file_found = True
                    if component not in self.required_components:
                        ChopprShares.log.debug(f"Component required: {component.name}-{component.version}")
                        self.required_components.add(component)
                        required_packages.update(self.component_packages[component])
                    break

            if not file_found:
                ChopprShares.log.error(f"Unable to determine what component provides file: {file}")

        ChopprShares.log.info(f"Found {len(self.required_components)} required DEB components")

        with self.cache_dir.joinpath("required-packages.txt").open("w", encoding="utf-8") as output:
            output.writelines([f"{pkg}\n" for pkg in sorted(required_packages, key=lambda p: p.name)])

    def populate_dependency_components(self) -> None:
        """Get all nested dependencies of the required components."""
        # Populate the dependencies from the dependencies section of the SBOM
        for component in self.required_components:
            ChopprShares.log.debug(f"Getting SBOM dependencies for {component.name}-{component.version}")
            amount_before = len(self.dependency_components)
            self._populate_nested_sbom_dependencies(component)
            ChopprShares.log.debug(f"Found {len(self.dependency_components) - amount_before} new SBOM dependencies")

        sbom_dependency_count = len(self.dependency_components)
        ChopprShares.log.info(f"Found {sbom_dependency_count} dependencies from the SBOM")

        # Check dependencies found in the package metadata
        for component in {
            c for c in self.required_components | self.dependency_components if get_purl_type(c) is self.purl_type
        }:
            ChopprShares.log.debug(f"Getting DEB dependencies for {component.name}-{component.version}")
            amount_before = len(self.dependency_components)
            self._populate_nested_file_dependencies(component)
            ChopprShares.log.debug(f"Found {len(self.dependency_components) - amount_before} new DEB dependencies")

        ChopprShares.log.info(f"Found {len(self.dependency_components) - sbom_dependency_count} DEB dependencies")

        dependency_packages = {
            package for component in self.dependency_components for package in self.component_packages[component]
        }

        with self.cache_dir.joinpath("dependency-packages.txt").open("w", encoding="utf-8") as file:
            file.writelines([f"{pkg}\n" for pkg in sorted(dependency_packages, key=lambda p: p.name)])

    ####################################################################################################
    # Utility Methods
    ####################################################################################################

    def _get_component_packages(self, component: Component) -> set[DebPackageData]:
        packages: set[DebPackageData] = set()
        if self.repositories:
            package_datas: set[DebPackageData] = {DebPackageData(package) for package in self.repositories.packages}
            packages.update({package for package in package_datas if package == component})

        return packages

    def _populate_nested_sbom_dependencies(self, component: Component) -> None:
        for dependency in get_component_dependencies(component):
            if dependency not in self.required_components | self.dependency_components:
                self.dependency_components.add(dependency)
                self._populate_nested_sbom_dependencies(dependency)

    @limit_recursion()
    def _populate_nested_file_dependencies(self, component: Component) -> None:
        new_dependencies: set[Component] = set()
        package_dependencies = {
            dependency
            for package_data in self.component_packages[component]
            for dependency in package_data.package.depends
        }
        unmatched_components = self.component_packages.keys() - (self.required_components | self.dependency_components)

        for package_dependency in package_dependencies:
            for unmatched_component in unmatched_components:
                if any(
                    package.satisfies_dependency(package_dependency)
                    for package in self.component_packages[unmatched_component]
                ) or any(
                    package.satisfies_dependency(alternate)
                    for package in self.component_packages[unmatched_component]
                    for alternate in package_dependency.alternates
                ):
                    self.dependency_components.add(unmatched_component)
                    new_dependencies.add(unmatched_component)

        for dependency in new_dependencies:
            self._populate_nested_sbom_dependencies(dependency)

        for dependency in new_dependencies:
            self._populate_nested_file_dependencies(dependency)



"""Class definition for PurlHandler."""

from __future__ import annotations

import abc

from typing import TYPE_CHECKING

from choppr.types.choppr_shares import ChopprShares


if TYPE_CHECKING:
    from pathlib import Path

    from hoppr import Component
    from hoppr.models.types import PurlType


__all__ = ["PurlHandler"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2025 Lockheed Martin Corporation"
__license__ = "MIT License"


class PurlHandler(abc.ABC):
    """Base class for handling package repositories of different PURL types."""

    def __init__(self, purl_type: PurlType) -> None:
        """Initialize the PurlHandler with a specific PURL type.

        Arguments:
            purl_type: The type of package URL (e.g., RPM, NPM, PyPI).
        """
        self.purl_type: PurlType = purl_type
        self.cache_dir: Path = ChopprShares.config.cache_dir / purl_type.value.lower()
        self.cache_dir.mkdir(parents=True, exist_ok=True)

        self.required_components: set[Component] = set()
        self.dependency_components: set[Component] = set()

    @abc.abstractmethod
    def cache_repositories(self) -> bool:
        """Pull metadata for repositories of the specified PURL type.

        Returns:
            bool: True if any repository was successfully pulled, False otherwise.
        """

    @abc.abstractmethod
    def resolve_component_packages(self) -> None:
        """Resolve all components to their packages in the repositories."""

    @abc.abstractmethod
    def populate_required_components(self, files: set[str]) -> None:
        """Determine which components provide the given files.

        Arguments:
            files: The files to search for in the component packages
        """

    @abc.abstractmethod
    def populate_dependency_components(self) -> None:
        """Get all nested dependencies of the required components."""



"""Class definition for ChopprConfig."""

from __future__ import annotations

import re

from datetime import timedelta
from enum import Enum
from hashlib import sha256
from pathlib import Path
from typing import Any

from hoppr import Component, PurlType
from pydantic import (  # noqa: TC002
    BaseModel,
    Field,
    HttpUrl,
    PositiveFloat,
    PositiveInt,
    PrivateAttr,
    parse_obj_as,
    validator,
)

from choppr import strace
from choppr.constants import ARCHIVE_EXTENSIONS, COMPONENT_LIST_FORMATS, DEFAULT_RECURSION_LIMIT


__all__ = ["ChopprConfig", "ChopprConfigModel", "DebianDistribution", "DebianRepository", "OperatingMode"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


def _validate_file_path(value: Path) -> Path:
    if not value.is_file():
        raise ValueError(f"File does not exist: {value}")
    return value


def _validate_http_url(value: str) -> HttpUrl:
    return parse_obj_as(HttpUrl, value.rstrip("/"))  # type: ignore[no-any-return]


def _validate_regex(value: str) -> re.Pattern[str]:
    try:
        return re.compile(value)
    except re.error as e:
        raise ValueError(f"Invalid regular expression: {value}") from e


def _default_excluded_components() -> dict[PurlType, ExcludedComponentsFile]:
    return {
        purl_type: ExcludedComponentsFile(
            file=Path(f"choppr-excluded-components-{purl_type.name.lower()}.txt"),
            component_format=COMPONENT_LIST_FORMATS.get(purl_type, "{name}={version}"),
        )
        for purl_type in PurlType
    }


class Certificate(BaseModel):
    # Required Attributes
    url: str
    certificate: Path

    # Validators
    _validate_certificate = validator("certificate", allow_reuse=True)(_validate_file_path)


class DebianDistribution(BaseModel):
    """Class representation for a debian distribution, to include its name and components.

    Members:
        - name
        - components
    """

    name: str
    components: list[str] = ["main", "restricted", "universe", "multiverse"]


class DebianRepository(BaseModel):
    """Class representation for a debian repository, to include its URL and distributions.

    Members:
        - url
        - distributions
    """

    url: HttpUrl
    distributions: list[DebianDistribution]

    _validate_url = validator("url", pre=True, allow_reuse=True)(_validate_http_url)


class HttpRequestLimits(BaseModel):
    """Class with values to configure HTTP request limits.

    Members:
        - retries
        - retry_interval
        - timeout
    """

    retries: PositiveInt = 3
    retry_interval: PositiveFloat = 5.0
    timeout: PositiveFloat = 60.0


class ExcludedComponentsFile(BaseModel):
    """The filename to output excluded components to, and what format to write them as.

    Members:
        - file
        - format
    """

    file: Path
    component_format: str = ""


class OutputFiles(BaseModel):
    """Class with values for the output files for Choppr.

    Members:
        - cache_archive
        - excluded_components
    """

    cache_archive: Path = Field(default_factory=lambda: Path.cwd() / "choppr-cache.tar.gz")
    excluded_components: dict[PurlType, ExcludedComponentsFile] = Field(default_factory=_default_excluded_components)

    @validator("cache_archive")
    @classmethod
    def _validate_cache_archive(cls, value: Path | None) -> Path | None:
        if value and value.suffix not in ARCHIVE_EXTENSIONS:
            raise ValueError(
                f"Invalid cache_archive value: {value} - Accepted extensions: {', '.join(ARCHIVE_EXTENSIONS)}"
            )

        return value

    @validator("excluded_components", pre=True)
    @classmethod
    def _validate_excluded_components(cls, value: dict[str, dict[str, str]]) -> dict[PurlType, ExcludedComponentsFile]:
        excluded_components: dict[PurlType, ExcludedComponentsFile] = _default_excluded_components()
        try:
            for purl, file_and_format in value.items():
                purl_type = PurlType[purl.upper()]
                excluded_components_file = ExcludedComponentsFile.parse_obj(file_and_format)
                excluded_components[purl_type].file = excluded_components_file.file
                if excluded_components_file.component_format:
                    excluded_components[purl_type].component_format = excluded_components_file.component_format
        except KeyError as e:
            raise ValueError(
                f"Invalid purl type: {e} - Accpeted values: [{', '.join(m.name for m in PurlType)}]"
            ) from e
        else:
            return excluded_components


class PackagePattern(BaseModel):
    """Class with the name, version, and purl type for a package.

    Members:
        - name
        - version
    """

    name: re.Pattern  # type: ignore[type-arg] # Pydantic v1 doesn't properly support re.Pattern
    version: re.Pattern  # type: ignore[type-arg] # Pydantic v1 doesn't properly support re.Pattern

    _validate_patterns = validator("name", "version", pre=True, allow_reuse=True)(_validate_regex)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, PackagePattern):
            return self.name == other.name and self.version == other.version
        if isinstance(other, Component):
            return bool(
                self.name.match(other.name)
                and (
                    (other.version is None and self.version.match(""))
                    or (other.version and self.version.match(other.version))
                )
            )
        return False

    def __hash__(self) -> int:
        sha = sha256()
        sha.update(str(self.name).encode())
        sha.update(str(self.version).encode())
        return int(sha.hexdigest(), 16)


class OperatingMode(str, Enum):
    """Modes to change the behavior of Choppr."""

    RUN = "run"
    """Run choppr normally, and filter the SBOM"""
    CACHE = "cache"
    """Only create the cache for Choppr, and export it as an archive."""


class ChopprConfigModel(BaseModel):
    """Class to validate and parse the configuration values provided to ChopprPlugin.

    Members:
        - strace_results
        - add_unmatched_file_components
        - allow_partial_filename_match
        - allow_version_mismatch
        - allowlist
        - archive_cache
        - cache_dir
        - cache_input
        - cache_timeout
        - certificates
        - clear_cache
        - deb_repositories
        - delete_excluded
        - denylist
        - http_limits
        - keep_essential_os_components
        - mode
        - output_files
        - recursion_limit
        - strace_regex_excludes

    Methods:
        - strace_files
    """

    HttpRequestLimits.update_forward_refs()

    # Operating mode
    mode: OperatingMode = OperatingMode.RUN

    # Required Attributes [RUN]
    strace_results: Path | None = None
    # Optional Attributes
    allow_partial_filename_match: bool = False
    allow_version_mismatch: bool = False
    allowlist: dict[PurlType, set[PackagePattern]] = Field(default={})
    archive_cache: bool = False
    cache_dir: Path = Field(default_factory=lambda: Path.cwd() / ".cache" / "choppr")
    cache_input: Path | None = None
    cache_timeout: timedelta | bool = Field(default=timedelta(days=7))
    certificates: dict[str, Path] = Field(default={})
    clear_cache: bool = False
    deb_repositories: list[DebianRepository] = Field(default=[])
    delete_excluded: bool = True
    denylist: dict[PurlType, set[PackagePattern]] = Field(default={})
    http_limits: HttpRequestLimits = Field(default=HttpRequestLimits())
    keep_essential_os_components: bool = False
    output_files: OutputFiles = Field(default=OutputFiles())
    recursion_limit: PositiveInt = DEFAULT_RECURSION_LIMIT
    strace_regex_excludes: list[re.Pattern] = Field(default=[])  # type: ignore[type-arg] # Pydantic v1 doesn't properly support re.Pattern
    # Private Attributes
    _strace_files: set[str] = PrivateAttr(default=set())

    # Validators
    _validate_strace_regex_excludes = validator("strace_regex_excludes", pre=True, each_item=True, allow_reuse=True)(
        _validate_regex
    )

    @validator("strace_results")
    @classmethod
    def _validate_strace_results(cls, value: Path, values: dict[str, Any]) -> Path | None:
        match values.get("mode"):
            case OperatingMode.RUN:
                if value is None:
                    raise ValueError("Invalid configuration: 'strace_results' is required in run mode")
                return _validate_file_path(value)
            case OperatingMode.CACHE:
                return None
        return None

    @validator("cache_input")
    @classmethod
    def _validate_cache_input(cls, value: Path | None) -> Path | None:
        if value:
            if value.suffix in ARCHIVE_EXTENSIONS:
                return _validate_file_path(value)
            raise ValueError(
                f"Invalid cache_input value: {value} - Accepted extensions: {', '.join(ARCHIVE_EXTENSIONS)}"
            )
        return value

    @validator("cache_timeout", pre=True)
    @classmethod
    def _validate_cache_timeout(cls, cache_timeout: str) -> timedelta | bool:
        timedelta_pattern = re.compile(r"^(?P<duration>\d+)\s?(?P<unit>d|h|m|s)$", re.IGNORECASE)
        boolean_pattern = re.compile(r"^(?P<true>true)|(?P<false>false)$", re.IGNORECASE)

        timeout_match = timedelta_pattern.match(cache_timeout)
        boolean_match = boolean_pattern.match(cache_timeout)

        error_message = "Invalid 'cache_timeout' value: Expected a number followed by a unit (d, h, m, s) or boolean"

        if not (timeout_match or boolean_match):
            raise ValueError(error_message)

        if boolean_match:
            return False if boolean_match["false"] else timedelta()

        timeout_duration = int(timeout_match["duration"])  # type: ignore[index]
        timeout_unit = timeout_match["unit"].lower()  # type: ignore[index]

        unit_map = {
            "d": "days",
            "h": "hours",
            "m": "minutes",
            "s": "seconds",
        }

        if timeout_unit not in unit_map:
            raise ValueError(error_message)

        return timedelta(**{unit_map[timeout_unit]: timeout_duration})

    @validator("certificates", pre=True)
    @classmethod
    def _validate_certificates(cls, certificates: list[dict[str, str]]) -> dict[str, Path]:
        certificate_map: dict[str, Path] = {}

        for certificate in certificates:
            c = Certificate.parse_obj(certificate)
            certificate_map[c.url] = c.certificate

        return certificate_map

    @validator("allowlist", "denylist", pre=True)
    @classmethod
    def _validate_exception_list(cls, value: dict[str, set[PackagePattern]]) -> dict[PurlType, set[PackagePattern]]:
        try:
            return {PurlType[purl_type.upper()]: packages for purl_type, packages in value.items()}
        except KeyError as e:
            raise ValueError(
                f"Invalid purl type: {e} - Accpeted values: [{', '.join(m.name for m in PurlType)}]"
            ) from e

    @validator("denylist")
    @classmethod
    def _validate_exception_overlap(
        cls, denylist: dict[PurlType, set[PackagePattern]], values: dict[str, Any]
    ) -> dict[PurlType, set[PackagePattern]]:
        allowlist: dict[PurlType, set[PackagePattern]] = parse_obj_as(
            dict[PurlType, set[PackagePattern]], values.get("allowlist")
        )

        for purl_type in PurlType:
            if allowlist.get(purl_type, set()) & denylist.get(purl_type, set()):
                raise ValueError(f"The allowlist and denylist have at least one overlapping {purl_type.name} package")

        return denylist

    def strace_files(self) -> set[str]:
        """List of files parsed from the provided strace results.

        Returns:
            set[Path]: List of files found from strace
        """
        if self.strace_results and not self._strace_files:
            parsed_strace_files = strace.get_files(self.strace_results)

            if self.strace_regex_excludes:
                self._strace_files = {
                    file
                    for file in parsed_strace_files
                    if not any(bool(re.search(exclude, str(file))) for exclude in self.strace_regex_excludes)
                }
                self.cache_dir.mkdir(parents=True, exist_ok=True)
                with self.cache_dir.joinpath("filtered-strace-results.txt").open("w") as output:
                    output.writelines([f"{file}\n" for file in self._strace_files])
            else:
                self._strace_files = parsed_strace_files

        return self._strace_files


class ChopprConfig:
    """A class to store the Choppr configuration.

    Members:
        - add_unmatched_file_components
        - allow_partial_filename_match
        - allow_version_mismatch
        - allowlist
        - archive_cache
        - cache_dir
        - cache_input
        - cache_timeout
        - certificates
        - clear_cache
        - deb_repositories
        - delete_excluded
        - denylist
        - http_limits
        - keep_essential_os_components
        - mode
        - output_files
        - recursion_limit
        - strace_files
    """

    def __init__(self, model: ChopprConfigModel) -> None:
        self.allow_partial_filename_match: bool = model.allow_partial_filename_match
        self.allow_version_mismatch: bool = model.allow_version_mismatch
        self.allowlist: dict[PurlType, set[PackagePattern]] = model.allowlist
        self.archive_cache: bool = model.archive_cache
        self.cache_dir: Path = model.cache_dir
        self.cache_input: Path | None = model.cache_input
        self.cache_timeout: timedelta | bool = model.cache_timeout
        self.certificates: dict[str, Path] = model.certificates
        self.clear_cache: bool = model.clear_cache
        self.deb_repositories: list[DebianRepository] = model.deb_repositories
        self.delete_excluded: bool = model.delete_excluded
        self.denylist: dict[PurlType, set[PackagePattern]] = model.denylist
        self.http_limits: HttpRequestLimits = model.http_limits
        self.keep_essential_os_components: bool = model.keep_essential_os_components
        self.mode: OperatingMode = model.mode
        self.output_files: OutputFiles = model.output_files
        self.recursion_limit: PositiveInt = model.recursion_limit
        self.strace_files: set[str] = model.strace_files()



"""Module for handling RPM packages."""

from __future__ import annotations

import gzip
import json
import lzma

from http import HTTPStatus
from typing import TYPE_CHECKING, Any, Final, cast

import jmespath
import xmltodict

from hoppr.models.manifest import Repository
from hoppr.models.types import PurlType
from pydantic import HttpUrl, parse_obj_as

from choppr.component_handlers.purl_handler import PurlHandler
from choppr.constants import DEFAULT_ARCH_RPM
from choppr.decorators import limit_recursion
from choppr.types import ChopprShares, RpmPackageData
from choppr.utils import (
    HTTP,
    cache_file_outdated,
    get_auth_and_verify,
    get_component_dependencies,
    get_purl_type,
    log_repo_pulls,
)


if TYPE_CHECKING:
    from collections import OrderedDict
    from pathlib import Path

    from hoppr import Component


__all__ = ["RpmHandler"]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class RpmHandler(PurlHandler):
    """Class to handle all RPM repository processing."""

    def __init__(self) -> None:
        super().__init__(PurlType.RPM)

        self.repositories: dict[str, OrderedDict[str, Any]] = {}
        self.component_packages: dict[Component, set[RpmPackageData]] = {}

    ####################################################################################################
    # Exported Methods
    ####################################################################################################

    def cache_repositories(self) -> bool:
        """Pull all of the metadata for RPM repositories provided in the config.

        Returns:
            bool: True when any repository was successfully pulled
        """
        if rpm_repository_urls := [
            parse_obj_as(HttpUrl, str(repo.url))
            for repo in parse_obj_as(list[Repository], ChopprShares.context.repositories[self.purl_type])
        ]:
            expected = len(rpm_repository_urls)
            ChopprShares.log.info(f"Caching {expected} RPM repositories...")

            for repo_url in rpm_repository_urls:
                if repo_return := self._get_remote_repo(repo_url):
                    repo_checksum, repo_data = repo_return
                    self.repositories[repo_checksum] = repo_data
            log_repo_pulls(expected, len(self.repositories), self.purl_type)

        return bool(len(self.repositories))

    def resolve_component_packages(self) -> None:
        """Resolve all RPM components to the packages in the repositories."""
        ChopprShares.log.info(
            f"Resolving packages for {len(ChopprShares.purl_components[self.purl_type])} RPM components"
        )
        for component in ChopprShares.purl_components[self.purl_type]:
            component_id = f"{component.name}-{component.version}"
            if packages := self._get_component_packages(component):
                if len(packages) == 1:
                    ChopprShares.log.debug(f"Resolved the package for {component_id}")
                else:
                    ChopprShares.log.warning(f"Resolved {len(packages)} packages for {component_id}")
                    for idx, package in enumerate(packages):
                        ChopprShares.log.warning(f"{idx}) {package}", indent_level=1)
            else:
                ChopprShares.log.warning(f"Unable to resolve a package for {component_id}")

            self.component_packages[component] = packages

    def populate_required_components(self, files: set[str]) -> None:
        """Determine which components provide the given files.

        Arguments:
            files: The files to search for in the component packages
        """
        required_packages: set[RpmPackageData] = set()

        for file in files:
            file_found = False
            for component in ChopprShares.purl_components[self.purl_type]:
                if any(package.provides_file(file, True) for package in self.component_packages[component]):
                    file_found = True
                    if component not in self.required_components:
                        ChopprShares.log.debug(f"Component required: {component.name}-{component.version}")
                        self.required_components.add(component)
                        required_packages.update(self.component_packages[component])

            if not file_found:
                ChopprShares.log.error(f"Unable to determine what component provides file: {file}")

        ChopprShares.log.info(f"Found {len(self.required_components)} required RPM components")

        with self.cache_dir.joinpath("required-packages.txt").open("w", encoding="utf-8") as output:
            output.writelines([f"{pkg}\n" for pkg in sorted(required_packages, key=lambda p: p.name)])

    def populate_dependency_components(self) -> None:
        """Get all nested dependencies of the required components."""
        # Populate the dependencies from the dependencies section of the SBOM
        for component in self.required_components:
            ChopprShares.log.debug(f"Getting SBOM dependencies for {component.name}-{component.version}")
            amount_before = len(self.dependency_components)
            self._populate_nested_sbom_dependencies(component)
            ChopprShares.log.debug(f"Found {len(self.dependency_components) - amount_before} new SBOM dependencies")

        sbom_dependency_count = len(self.dependency_components)
        ChopprShares.log.info(f"Found {sbom_dependency_count} dependencies from the SBOM")

        # Check dependencies found in the package metadata
        for component in {
            c for c in self.required_components | self.dependency_components if get_purl_type(c) is self.purl_type
        }:
            ChopprShares.log.debug(f"Getting RPM dependencies for {component.name}-{component.version}")
            amount_before = len(self.dependency_components)
            self._populate_nested_file_dependencies(component)
            ChopprShares.log.debug(f"Found {len(self.dependency_components) - amount_before} new RPM dependencies")

        ChopprShares.log.info(f"Found {len(self.dependency_components) - sbom_dependency_count} RPM dependencies")

        dependency_packages = {
            package for component in self.dependency_components for package in self.component_packages[component]
        }

        with self.cache_dir.joinpath("dependency-packages.txt").open("w", encoding="utf-8") as file:
            file.writelines([f"{pkg}\n" for pkg in sorted(dependency_packages, key=lambda p: p.name)])

    ####################################################################################################
    # Utility Methods
    ####################################################################################################

    def _get_remote_repo(self, repo_url: HttpUrl) -> tuple[str, OrderedDict[str, Any]] | None:
        repository_dir: Final[Path] = self.cache_dir.joinpath("repositories", f"{repo_url.host}{repo_url.path or ''}")
        repository_dir.mkdir(parents=True, exist_ok=True)

        cache_repomd = repository_dir / "repomd.json"
        repomd_url = parse_obj_as(HttpUrl, f"{repo_url}/repodata/repomd.xml")
        basic_auth, verify = get_auth_and_verify(repomd_url)

        if not cache_file_outdated(cache_repomd):
            ChopprShares.log.info(f"Loading repository data from {cache_repomd}...")
            with cache_repomd.open() as file:
                repomd_data = cast("OrderedDict[str, Any]", json.load(file))
        else:
            ChopprShares.log.info(f"Pulling repository data from {repomd_url}...")
            with HTTP.get(repomd_url, basic_auth, verify) as response:
                if response.status_code > HTTPStatus.MULTIPLE_CHOICES:
                    ChopprShares.log.error(f"Failed to pull repository: {response.content.decode()}")
                    return None

                repomd_data = xmltodict.parse(response.content)

                with cache_repomd.open("w") as file:
                    json.dump(repomd_data, file, indent=2)

        primary_checksum = jmespath.search(
            expression='[?"@type"==\'primary\'].checksum."#text"',
            data=repomd_data["repomd"]["data"],
        )
        primary_location = jmespath.search(
            expression='[?"@type"==\'primary\'].location."@href"',
            data=repomd_data["repomd"]["data"],
        )

        if isinstance(primary_checksum, list) and len(primary_checksum) == 1:
            cache_repository_file = repository_dir / f"{primary_checksum[0]}.json"

            # Use cached repository
            if cache_repository_file.is_file():
                with cache_repository_file.open() as file:
                    ChopprShares.log.info("Successfully loaded repository from cache")
                    return (primary_checksum[0], cast("OrderedDict[str, Any]", json.load(file)))
            # Pull repository
            elif isinstance(primary_location, list) and len(primary_location) == 1:
                primary_url = parse_obj_as(HttpUrl, f"{repo_url}/{primary_location[0]}")
                with HTTP.get(primary_url, basic_auth, verify) as primary_response:
                    if primary_response.status_code > HTTPStatus.MULTIPLE_CHOICES:
                        ChopprShares.log.error(f"Failed to pull repository: {response.content.decode()}")
                        return None

                    match primary_url:
                        case _ if primary_url.endswith(".gz"):
                            primary_xml = gzip.decompress(primary_response.content)
                        case _ if primary_url.endswith(".xz"):
                            primary_xml = lzma.decompress(primary_response.content)
                        case _ if primary_url.endswith(".xml"):
                            primary_xml = primary_response.content
                        case _:
                            ChopprShares.log.error(f"Unsupported file type found for primary repository: {primary_url}")
                            return None

                    ChopprShares.log.info("Successfully pulled repository data")

                    repo_data = xmltodict.parse(primary_xml)
                    # Write repository data to cache
                    with cache_repository_file.open("w") as file:
                        json.dump(repo_data, file, indent=2)
                    return (primary_checksum[0], repo_data)

        ChopprShares.log.error("Failed to get `location` and/or `checksum` from repository")
        return None

    def _get_component_packages(self, component: Component, arch: str = DEFAULT_ARCH_RPM) -> set[RpmPackageData]:
        packages: set[RpmPackageData] = set()
        for repo in self.repositories.values():
            package_datas: set[RpmPackageData] = {
                RpmPackageData(package)
                for package in jmespath.search(
                    expression=f"metadata.package[?arch == '{arch}' || arch == 'noarch']", data=repo
                )
            }

            packages.update({package for package in package_datas if package == component})

        return packages

    def _populate_nested_sbom_dependencies(self, component: Component) -> None:
        for dependency in get_component_dependencies(component):
            if dependency not in self.required_components | self.dependency_components:
                self.dependency_components.add(dependency)
                self._populate_nested_sbom_dependencies(dependency)

    @limit_recursion()
    def _populate_nested_file_dependencies(self, component: Component) -> None:
        new_dependencies: set[Component] = set()
        required_files = {file for package in self.component_packages[component] for file in package.requires()}
        unmatched_components = self.component_packages.keys() - (self.required_components | self.dependency_components)

        for file in required_files:
            for unmatched_component in unmatched_components:
                if any(package.provides_file(file) for package in self.component_packages[unmatched_component]):
                    self.dependency_components.add(unmatched_component)
                    new_dependencies.add(unmatched_component)

        for dependency in new_dependencies:
            self._populate_nested_sbom_dependencies(dependency)

        for dependency in new_dependencies:
            self._populate_nested_file_dependencies(dependency)



"""Utility functions used within the Choppr plugin."""

from __future__ import annotations

import bz2
import contextlib
import gzip
import lzma
import re
import tarfile
import time
import zipfile

from datetime import datetime
from http.client import RemoteDisconnected
from typing import TYPE_CHECKING

import requests

from hoppr import Credentials, PurlType
from pydantic import HttpUrl, SecretStr, parse_obj_as
from requests.auth import HTTPBasicAuth
from requests.exceptions import ProxyError

from choppr.constants import LOG_HEADER_LENGTH
from choppr.types import ChopprShares
from choppr.types.choppr_config import OperatingMode


if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from hoppr import Component
    from requests import Response


__all__ = [
    "HTTP",
    "cache_file_outdated",
    "compress_directory",
    "extract_archive",
    "filenames_match",
    "get_auth_and_verify",
    "get_component_dependencies",
    "get_component_from_ref",
    "get_purl_type",
    "log_header",
    "log_repo_pulls",
    "output_list",
    "prepend_slash",
    "remove_parenthesis",
]
__author__ = "Jesse Boswell <jesse.a.boswell@lmco.com>"
__copyright__ = "Copyright (C) 2024 Lockheed Martin Corporation"
__license__ = "MIT License"


class HTTP:
    """Utility methods to handle HTTP requests."""

    @staticmethod
    def get(url: HttpUrl, auth: HTTPBasicAuth | None = None, verify: str | bool = True) -> Response:
        """Perform an HTTP get request with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)

        Returns:
            Response: The response from the request
        """
        limits = ChopprShares.config.http_limits

        for attempt in range(limits.retries):
            if attempt > 0:
                time.sleep(limits.retry_interval)

            with contextlib.suppress(RemoteDisconnected):
                response = requests.get(
                    url,
                    auth=auth,
                    stream=True,
                    verify=verify,
                    timeout=limits.timeout,
                )
                if response.ok:
                    return response
        return response

    @staticmethod
    def download_raw(url: HttpUrl, auth: HTTPBasicAuth | None = None, verify: str | bool = True) -> bytes | None:
        """Download the raw content with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)

        Returns:
            bytes | None: The raw content if the request is successful
        """
        response = HTTP.get(url, auth, verify)

        return response.content if response.ok else None

    @staticmethod
    def download(
        url: HttpUrl,
        auth: HTTPBasicAuth | None = None,
        verify: str | bool = True,
        encoding: str = "utf-8",
    ) -> str | None:
        """Download the content and decode with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)
            encoding: The encoding to use when decoding the content (default "utf-8")

        Returns:
            str | None: The decoded content if the request is successful
        """
        with contextlib.suppress(ProxyError):
            if raw := HTTP.download_raw(url, auth, verify):
                return raw.decode(encoding)

        return None

    @staticmethod
    def download_compressed(
        url: HttpUrl,
        auth: HTTPBasicAuth | None = None,
        verify: str | bool = True,
        encoding: str = "utf-8",
    ) -> str | None:
        """Download the content, then decoompress and decode it with the provided parameters.

        Arguments:
            url: The URL to access
            auth: The credentials needed to access the URL (default None)
            verify: The certificate needed to access the URL (default True)
            encoding: The encoding to use when decoding the content (default "utf-8")

        Returns:
            str | None: The decompressed and decoded content if the request is successful
        """
        decompress: dict[str, Callable[[bytes], bytes]] = {
            "": lambda c: c,
            ".xz": lzma.decompress,
            ".gz": gzip.decompress,
            ".tgz": gzip.decompress,
            ".bzip2": bz2.decompress,
        }

        for suffix, method in decompress.items():
            with contextlib.suppress(ProxyError):
                if (raw := HTTP.download_raw(parse_obj_as(HttpUrl, f"{url}{suffix}"), auth, verify)) and (
                    output := method(raw).decode(encoding)
                ):
                    return output

        return None


def cache_file_outdated(file: Path) -> bool:
    """Check if the provided file is outdated.

    Arguments:
        file: File to check

    Returns:
        bool: True if the file is outdated, otherwise, false
    """
    if ChopprShares.config.mode == OperatingMode.CACHE:
        ChopprShares.log.debug(
            f"Running in cache mode. Refreshing file: {file.relative_to(ChopprShares.config.cache_dir)}"
        )
        return True

    if not file.is_file():
        ChopprShares.log.debug(
            f"Missing cache file will be downloaded: {file.relative_to(ChopprShares.config.cache_dir)}"
        )
        return True

    if isinstance(ChopprShares.config.cache_timeout, bool):
        ChopprShares.log.debug(f"Cache timeout is {'enabled' if ChopprShares.config.cache_timeout else 'disabled'}")
        return ChopprShares.config.cache_timeout

    tz = datetime.now().astimezone().tzinfo
    expiration_time = datetime.fromtimestamp(file.stat().st_mtime, tz) + ChopprShares.config.cache_timeout

    if outdated := expiration_time < datetime.now(tz):
        ChopprShares.log.debug(
            f"Outdated cache file will be refreshed: {file.relative_to(ChopprShares.config.cache_dir)}"
        )

    return outdated


def get_auth_and_verify(url: HttpUrl) -> tuple[HTTPBasicAuth | None, str | bool]:
    """Get the authentication credentials and certificate.

    Arguments:
        url: URL to get the credentials and certificate for

    Returns:
        tuple[HTTPBasicAuth | None, str | bool]: Credentials and certificate
    """
    if not url.host:
        return (None, True)

    auth = None
    credentials = Credentials.find(url.host)
    certificate = ChopprShares.config.certificates.get(url.host)
    if credentials and isinstance(credentials.password, SecretStr):
        ChopprShares.log.debug(f"Found credentials for host: {url.host}")
        auth = HTTPBasicAuth(username=credentials.username, password=credentials.password.get_secret_value())
    else:
        ChopprShares.log.debug(f"No credentials for host: {url.host}")

    return (auth, str(certificate) if certificate else True)


def get_purl_type(component: Component) -> PurlType | None:
    """Get the purl type for the provided component.

    Arguments:
        component: SBOM component

    Returns:
        PurlType | None: The purl type for the component
    """
    if component.purl and (match := re.match(r"pkg:(?P<type>.*?)/", component.purl)):
        with contextlib.suppress(KeyError):
            return PurlType(match["type"].upper())
    return None


def log_header(title: str) -> None:
    """Output a header to the log with the provided title, and padded with equal signs.

    Arguments:
        title: The header title
    """
    ChopprShares.log.info(f" {title} ".center(LOG_HEADER_LENGTH, "="))


def log_repo_pulls(expected: int, actual: int, purl_type: PurlType) -> None:
    """Log the success or failure of pulling a repository.

    Arguments:
        expected: The expected number of repositories
        actual: The number of repositories actually pulled
        purl_type: They repository type
    """
    match actual:
        case 0:
            ChopprShares.log.error(f"Failed to pull any {purl_type} repositories")
        case _ if actual == expected:
            ChopprShares.log.info(f"Successfully pulled all {purl_type} repositories")
        case _:
            ChopprShares.log.error(f"Failed to pull {expected - actual}/{expected} {purl_type} repositories")


def output_list(output_file: Path, items: list[str]) -> None:
    """Output list to the provided file.

    Arguments:
        output_file: Output file path
        items: List of strings to write to the file
    """
    with output_file.open("w", encoding="utf-8") as file:
        file.writelines([f"{item}\n" for item in items])


def prepend_slash(text: str) -> str:
    """Prepend a slash to a string that doesn't have one.

    Arguments:
        text: A string to prepend a slash to

    Returns:
        str: The string with a prepended slash
    """
    return f"/{text.removeprefix('/')}"


def remove_parenthesis(text: str) -> str:
    """Remove text within parenthesis, to include empty parenthesis.

    Arguments:
        text: Text to remove parenthesis from

    Returns:
        str: Text with parenthesis removed
    """
    return re.sub(r"\(.*?\)", "", text)


def compress_directory(archive: Path, input_dir: Path) -> None:
    """Compress all files in a given directory.

    Arguments:
        archive: The archive file to create
        input_dir: The directory to traverse and compress
    """
    match archive.suffix:
        case ".bz2":
            with tarfile.open(archive, "w:bz2") as archive_file:
                for file in input_dir.rglob("*"):
                    archive_file.add(file, file.relative_to(input_dir), False)
        case ".gz" | ".tgz":
            with tarfile.open(archive, "w:gz") as archive_file:
                for file in input_dir.rglob("*"):
                    archive_file.add(file, file.relative_to(input_dir), False)
        case ".xz":
            with tarfile.open(archive, "w:xz") as archive_file:
                for file in input_dir.rglob("*"):
                    archive_file.add(file, file.relative_to(input_dir), False)
        case ".zip":
            with zipfile.ZipFile(archive, "w", zipfile.ZIP_DEFLATED) as archive_file:
                for file in input_dir.rglob("*"):
                    archive_file.write(file, file.relative_to(input_dir))


def _is_safe_to_extract(info: tarfile.TarInfo | zipfile.ZipInfo, output_dir: Path) -> bool:
    archived_file = info.name if isinstance(info, tarfile.TarInfo) else info.filename

    return output_dir.joinpath(archived_file).resolve().is_relative_to(output_dir.resolve())


def extract_archive(archive: Path, output_dir: Path) -> None:
    """Extract the contents of the given archive to the output directory.

    Arguments:
        archive: The archive to extract
        output_dir: The output directory
    """
    match archive.suffix:
        case ".bz2":
            with tarfile.open(archive, "r:bz2") as archive_file:
                tar_members = [
                    member for member in archive_file.getmembers() if _is_safe_to_extract(member, output_dir)
                ]
                archive_file.extractall(output_dir, tar_members)  # noqa: S202
        case ".gz" | ".tgz":
            with tarfile.open(archive, "r:gz") as archive_file:
                tar_members = [
                    member for member in archive_file.getmembers() if _is_safe_to_extract(member, output_dir)
                ]
                archive_file.extractall(output_dir, tar_members)  # noqa: S202
        case ".xz":
            with tarfile.open(archive, "r:xz") as archive_file:
                tar_members = [
                    member for member in archive_file.getmembers() if _is_safe_to_extract(member, output_dir)
                ]
                archive_file.extractall(output_dir, tar_members)  # noqa: S202
        case ".zip":
            with zipfile.ZipFile(archive, "r", zipfile.ZIP_DEFLATED) as archive_file:
                zip_members = [member for member in archive_file.filelist if _is_safe_to_extract(member, output_dir)]
                archive_file.extractall(output_dir, zip_members)  # noqa: S202


def filenames_match(left: str, right: str, force_partial_match: bool = False) -> bool:
    """Compare the given filenames to check if they match.

    Substring matching will be used if the `allow_partial_filename_match` config option is set to True.

    Arguments:
        left: The filename to compare to
        right: The filename to compare against
        force_partial_match: Force partial matching of the filenames (default False)

    Returns:
        bool: True if the filenames match
    """
    if left == right:
        return True

    return (
        left in right or right in left
        if ChopprShares.config.allow_partial_filename_match or force_partial_match
        else False
    )


def get_component_dependencies(component: Component) -> set[Component]:
    """Check if the given component is in the dependencies section of an SBOM.

    Arguments:
        component: The component to get the dependencies of

    Returns:
        set[Component]: The list of dependencies for a component
    """
    dependencies: set[Component] = set()

    if ChopprShares.context.delivered_sbom.dependencies and (
        dependency := next(
            (d for d in ChopprShares.context.delivered_sbom.dependencies if d.ref == component.bom_ref), None
        )
    ):
        dependencies.update(
            ref_component for ref in (dependency.dependsOn or []) if (ref_component := get_component_from_ref(ref))
        )

    return dependencies


def get_component_from_ref(ref: str) -> Component | None:
    """Get the component from the SBOM that has a bom_ref that matches the given ref.

    Arguments:
        ref: The ref to search for

    Returns:
        Component | None: The component matching the given ref
    """
    return next((c for c in ChopprShares.context.delivered_sbom.components if c.bom_ref == ref), None)
